@@ -37,6 +37,7 @@ import io
 import json
 import os
 import itertools
+import codecs
 
 # Dependencies
 import urllib3
@@ -44,8 +45,10 @@ import certifi
 
 # Internal imports
 from . import __title__, __version__
-from .compat import urlsplit, range, urlencode, basestring
-from .error import ResponseError, InvalidArgumentError, InvalidSizeError
+from .compat import (urlsplit, queryencode,
+                     range, basestring)
+from .error import (KnownResponseError, ResponseError, NoSuchBucket,
+                    InvalidArgumentError, InvalidSizeError, NoSuchBucketPolicy)
 from .definitions import Object, UploadPart
 from .parsers import (parse_list_buckets,
                       parse_list_objects,
@@ -74,7 +77,6 @@ from .xml_marshal import (xml_marshal_bucket_constraint,
                           xml_marshal_complete_multipart_upload,
                           xml_marshal_bucket_notifications,
                           xml_marshal_delete_objects)
-from .limited_reader import LimitedReader
 from . import policy
 from .fold_case_dict import FoldCaseDict
 
@@ -116,10 +118,12 @@ class Minio(object):
          are allowed to run before being aborted.
     :return: :class:`Minio <Minio>` object
     """
+
     def __init__(self, endpoint, access_key=None,
                  secret_key=None, secure=True,
                  region=None,
-                 timeout=None):
+                 timeout=None,
+                 certificate_bundle=certifi.where()):
 
         # Validate endpoint.
         is_valid_endpoint(endpoint)
@@ -145,7 +149,7 @@ class Minio(object):
         self._http = urllib3.PoolManager(
             timeout=self._conn_timeout,
             cert_reqs='CERT_REQUIRED',
-            ca_certs=certifi.where(),
+            ca_certs=certificate_bundle,
             retries=urllib3.Retry(
                 total=5,
                 backoff_factor=0.2,
@@ -252,8 +256,7 @@ class Minio(object):
                                       headers=headers)
 
         if response.status != 200:
-            response_error = ResponseError(response)
-            raise response_error.put(bucket_name)
+            raise ResponseError(response, method, bucket_name).get_exception()
 
         self._set_bucket_region(bucket_name, region=location)
 
@@ -292,8 +295,7 @@ class Minio(object):
                       self._trace_output_stream)
 
         if response.status != 200:
-            response_error = ResponseError(response)
-            raise response_error.get()
+            raise ResponseError(response, method).get_exception()
 
         return parse_list_buckets(response.data)
 
@@ -309,11 +311,10 @@ class Minio(object):
         try:
             self._url_open('HEAD', bucket_name=bucket_name)
         # If the bucket has not been created yet, Minio will return a "NoSuchBucket" error.
+        except NoSuchBucket as e:
+            return False
         except ResponseError as e:
-            if e.code == 'NoSuchBucket':
-                return False
             raise
-
         return True
 
     def remove_bucket(self, bucket_name):
@@ -334,12 +335,10 @@ class Minio(object):
             response = self._url_open("GET",
                                       bucket_name=bucket_name,
                                       query={"policy": ""})
+        except NoSuchBucketPolicy as e:
+            return None
         except ResponseError as e:
-            # Ignore 'NoSuchBucketPolicy' error.
-            if e.code != 'NoSuchBucketPolicy':
-                raise
-            else:
-                return None
+            raise
 
         data = response.data
         if isinstance(data, bytes) and isinstance(data, str):  # Python 2
@@ -366,7 +365,7 @@ class Minio(object):
 
         # Normalize statements.
         statements = []
-        policy.append_statements(statements, policy_dict.get('Statement', []))
+        policy._append_statements(statements, policy_dict.get('Statement', []))
 
         return policy.get_policy(statements, bucket_name, prefix)
 
@@ -389,7 +388,7 @@ class Minio(object):
 
         # Normalize statements.
         statements = []
-        policy.append_statements(statements, policy_dict['Statement'])
+        policy._append_statements(statements, policy_dict['Statement'])
 
         statements = policy.set_policy(statements, policy_access,
                                        bucket_name, prefix)
@@ -482,7 +481,10 @@ class Minio(object):
             content_sha256=content_sha256_hex
         )
 
-    def listen_bucket_notification(self, bucket_name, prefix, suffix, events):
+    def listen_bucket_notification(self, bucket_name, prefix='', suffix='',
+                                   events=['s3:ObjectCreated:*',
+                                           's3:ObjectRemoved:*',
+                                           's3:ObjectAccessed:*']):
         """
         Yeilds new event notifications on a bucket, caller should iterate
         to read new notifications.
@@ -501,7 +503,7 @@ class Minio(object):
         url_components = urlsplit(self._endpoint_url)
         if url_components.hostname == 's3.amazonaws.com':
             raise InvalidArgumentError(
-                'Listening for event notifications on a  bucket is a Minio '
+                'Listening for event notifications on a bucket is a Minio '
                 'specific extension to bucket notification API. It is not '
                 'supported by Amazon S3')
 
@@ -512,7 +514,7 @@ class Minio(object):
         }
         while True:
             response = self._url_open('GET', bucket_name=bucket_name,
-                                      query=query)
+                                      query=query, preload_content=False)
             try:
                 for line in response.stream():
                     event = json.loads(line)
@@ -522,35 +524,6 @@ class Minio(object):
                 response.close()
                 continue
 
-    def _get_upload_id(self, bucket_name, object_name, metadata=None):
-        """
-        Get previously uploaded upload id for object name or initiate a request
-        to fetch a new upload id.
-
-        :param bucket_name: Bucket name where the incomplete upload resides.
-        :param object_name: Object name for which upload id is requested.
-        :param metadata: Additional metadata headers for new multipart upload.
-        """
-        recursive = True
-        current_uploads = self._list_incomplete_uploads(bucket_name,
-                                                        object_name,
-                                                        recursive,
-                                                        is_aggregate_size=False)
-        matching_uploads = [upload
-                            for upload in current_uploads
-                            if object_name == upload.object_name]
-
-        # If no matching uploads its a new multipart upload.
-        if not len(matching_uploads):
-            upload_id = self._new_multipart_upload(bucket_name,
-                                                   object_name,
-                                                   metadata)
-        else:
-            incomplete_upload = max(matching_uploads, key=lambda x: x.initiated)
-            upload_id = incomplete_upload.upload_id
-
-        return upload_id
-
     def fput_object(self, bucket_name, object_name, file_path,
                     content_type='application/octet-stream',
                     metadata=None):
@@ -568,125 +541,15 @@ class Minio(object):
             with your PUT request.
         :return: etag
         """
-        is_valid_bucket_name(bucket_name)
-        is_non_empty_string(object_name)
-        is_non_empty_string(file_path)
-
-        # save file_size.
-        file_size = os.stat(file_path).st_size
-
-        if file_size > MAX_MULTIPART_OBJECT_SIZE:
-            raise InvalidArgumentError('Input content size is bigger '
-                                       ' than allowed maximum of 5TiB.')
 
         # Open file in 'read' mode.
         file_data = io.open(file_path, mode='rb')
+        file_size = os.stat(file_path).st_size
 
-        if file_size <= MIN_OBJECT_SIZE:
-            return self._do_put_object(bucket_name, object_name,
-                                       file_data.read(file_size),
-                                       metadata=metadata)
+        return self.put_object(bucket_name, object_name, file_data, file_size,
+                content_type, metadata)
 
-        if not metadata:
-            metadata = {}
-
-        metadata['Content-Type'] = 'application/octet-stream' if \
-                                   not content_type else content_type
-
-        # Calculate optimal part info.
-        total_parts_count, part_size, last_part_size = optimal_part_info(file_size)
-
-        # get upload id.
-        upload_id = self._get_upload_id(bucket_name, object_name, metadata)
-
-        # Initialize variables
-        uploaded_parts = {}
-        total_uploaded = 0
-
-        # Iter over the uploaded parts.
-        parts_iter = self._list_object_parts(bucket_name,
-                                             object_name,
-                                             upload_id)
-
-        for part in parts_iter:
-            # Save uploaded parts for future verification.
-            uploaded_parts[part.part_number] = part
-
-        # Always start with first part number.
-        for part_number in range(1, total_parts_count + 1):
-            # Save the current part size that needs to be uploaded.
-            current_part_size = part_size
-            if part_number == total_parts_count:
-                current_part_size = last_part_size
-
-            # Save current offset as previous offset.
-            prev_offset = file_data.seek(0, 1)
-
-            # Calculate md5sum and sha256.
-            md5hasher = Hasher.md5()
-            sha256hasher = Hasher.sha256()
-            total_read = 0
-
-            # Save LimitedReader, read upto current_part_size for
-            # md5sum and sha256 calculation.
-            part = LimitedReader(file_data, current_part_size)
-            while total_read < current_part_size:
-                current_data = part.read() # Read in 64k chunks.
-                if not current_data or len(current_data) == 0:
-                    break
-                md5hasher.update(current_data)
-                sha256hasher.update(current_data)
-                total_read += len(current_data)
-
-            part_md5_hex = md5hasher.hexdigest()
-            # Verify if current part number has been already
-            # uploaded. Verify if the size is same, further verify if
-            # we have matching md5sum as well.
-            if part_number in uploaded_parts:
-                previous_part = uploaded_parts[part_number]
-                if previous_part.size == current_part_size:
-                    if previous_part.etag == part_md5_hex:
-                        total_uploaded += previous_part.size
-                        continue
-
-            # Seek back to previous offset position before checksum
-            # calculation.
-            file_data.seek(prev_offset, 0)
-
-            # Create the LimitedReader again for the http reader.
-            part = LimitedReader(file_data, current_part_size)
-            part_metadata = PartMetadata(part, md5hasher, sha256hasher,
-                                         current_part_size)
-            # Initiate multipart put.
-            etag = self._do_put_multipart_object(bucket_name, object_name,
-                                                 part_metadata,
-                                                 upload_id,
-                                                 part_number)
-
-            # Save etags.
-            uploaded_parts[part_number] = UploadPart(bucket_name,
-                                                     object_name,
-                                                     upload_id,
-                                                     part_number,
-                                                     etag, None,
-                                                     total_read)
-            # Total uploaded.
-            total_uploaded += total_read
-
-        if total_uploaded != file_size:
-            msg = 'Data uploaded {0} is not equal input size ' \
-                  '{1}'.format(total_uploaded, file_size)
-            raise InvalidSizeError(msg)
-
-        # Complete all multipart transactions if possible.
-        mpart_result = self._complete_multipart_upload(bucket_name,
-                                                       object_name,
-                                                       upload_id,
-                                                       uploaded_parts)
-        # Return etag here.
-        return mpart_result.etag
-
-    def fget_object(self, bucket_name, object_name, file_path):
+    def fget_object(self, bucket_name, object_name, file_path, request_headers=None):
         """
         Retrieves an object from a bucket and writes at file_path.
 
@@ -696,6 +559,7 @@ class Minio(object):
         :param bucket_name: Bucket to read object from.
         :param object_name: Name of the object to read.
         :param file_path: Local file path to save the object.
+        :param request_headers: Any additional headers to be added with GET request.
         """
         is_valid_bucket_name(bucket_name)
         is_non_empty_string(object_name)
@@ -721,14 +585,15 @@ class Minio(object):
             # Get partial object.
             response = self._get_partial_object(bucket_name, object_name,
                                                 offset=file_statinfo.st_size,
-                                                length=0)
+                                                length=0,
+                                                request_headers=request_headers)
 
             # Save content_size to verify if we wrote more data.
             content_size = int(response.headers['content-length'])
 
             # Save total_written.
             total_written = 0
-            for data in response.stream(amt=1024*1024):
+            for data in response.stream(amt=1024 * 1024):
                 file_part_data.write(data)
                 total_written += len(data)
 
@@ -754,7 +619,7 @@ class Minio(object):
         # Return the stat
         return stat
 
-    def get_object(self, bucket_name, object_name):
+    def get_object(self, bucket_name, object_name, request_headers=None):
         """
         Retrieves an object from a bucket.
 
@@ -769,6 +634,7 @@ class Minio(object):
 
         :param bucket_name: Bucket to read object from
         :param object_name: Name of object to read
+        :param request_headers: Any additional headers to be added with GET request.
         :return: :class:`urllib3.response.HTTPResponse` object.
 
         """
@@ -776,9 +642,10 @@ class Minio(object):
         is_non_empty_string(object_name)
 
         return self._get_partial_object(bucket_name,
-                                        object_name)
+                                        object_name,
+                                        request_headers=request_headers)
 
-    def get_partial_object(self, bucket_name, object_name, offset=0, length=0):
+    def get_partial_object(self, bucket_name, object_name, offset=0, length=0, request_headers=None):
         """
         Retrieves an object from a bucket.
 
@@ -799,6 +666,7 @@ class Minio(object):
            Must be >= 0.
         :param length: Optional number of bytes to retrieve.
            Must be an integer.
+        :param request_headers: Any additional headers to be added with GET request.
         :return: :class:`urllib3.response.HTTPResponse` object.
 
         """
@@ -807,7 +675,8 @@ class Minio(object):
 
         return self._get_partial_object(bucket_name,
                                         object_name,
-                                        offset, length)
+                                        offset, length,
+                                        request_headers=request_headers)
 
     def copy_object(self, bucket_name, object_name, object_source,
                     conditions=None):
@@ -832,7 +701,7 @@ class Minio(object):
         if conditions:
             headers = {k: v for k, v in conditions.items()}
 
-        headers['X-Amz-Copy-Source'] = urlencode(object_source)
+        headers['X-Amz-Copy-Source'] = queryencode(object_source)
         response = self._url_open('PUT',
                                   bucket_name=bucket_name,
                                   object_name=object_name,
@@ -870,7 +739,8 @@ class Minio(object):
         is_valid_bucket_name(bucket_name)
         is_non_empty_string(object_name)
         if not callable(getattr(data, 'read')):
-            raise ValueError('Invalid input data does not implement a callable read() method')
+            raise ValueError(
+                'Invalid input data does not implement a callable read() method')
 
         if length > MAX_MULTIPART_OBJECT_SIZE:
             raise InvalidArgumentError('Input content size is bigger '
@@ -879,7 +749,7 @@ class Minio(object):
             metadata = {}
 
         metadata['Content-Type'] = 'application/octet-stream' if \
-                                   not content_type else content_type
+            not content_type else content_type
 
         if length > MIN_OBJECT_SIZE:
             return self._stream_put_object(bucket_name, object_name,
@@ -936,8 +806,9 @@ class Minio(object):
 
         # Initialize query parameters.
         query = {
-            'max-keys': 1000
+            'max-keys': '1000'
         }
+
         # Add if prefix present.
         if prefix:
             query['prefix'] = prefix
@@ -1002,7 +873,7 @@ class Minio(object):
 
         # Initialize query parameters.
         query = {
-            'list-type': 2
+            'list-type': '2'
         }
         # Add if prefix present.
         if prefix:
@@ -1045,11 +916,26 @@ class Minio(object):
         size = int(response.headers.get('content-length', '0'))
         content_type = response.headers.get('content-type', '')
         last_modified = response.headers.get('last-modified')
+
+        ## Supported headers for object.
+        supported_headers = [
+            'cache-control',
+            'content-encoding',
+            'content-disposition',
+            ## Add more supported headers here.
+        ]
+
+        ## Capture only custom metadata.
+        custom_metadata = dict()
+        for k in response.headers:
+            if k in supported_headers or k.lower().startswith('x-amz-meta-'):
+                custom_metadata[k] = response.headers.get(k)
+
         if last_modified:
             http_time_format = "%a, %d %b %Y %H:%M:%S GMT"
             last_modified = mktime(strptime(last_modified, http_time_format))
-        return Object(bucket_name, object_name, content_type=content_type,
-                      last_modified=last_modified, etag=etag, size=size)
+        return Object(bucket_name, object_name, last_modified, etag, size,
+                      content_type=content_type, metadata=custom_metadata)
 
     def remove_object(self, bucket_name, object_name):
         """
@@ -1079,7 +965,7 @@ class Minio(object):
             'Content-Md5': get_md5_base64digest(content),
             'Content-Length': len(content)
         }
-        query = {"delete": None}
+        query = {'delete': ''}
         content_sha256_hex = get_sha256_hexdigest(content)
 
         # send multi-object delete request
@@ -1188,7 +1074,8 @@ class Minio(object):
 
         return self._list_incomplete_uploads(bucket_name, prefix, recursive)
 
-    def _list_incomplete_uploads(self, bucket_name, prefix=None, recursive=False, is_aggregate_size=True):
+    def _list_incomplete_uploads(self, bucket_name, prefix=None,
+                                 recursive=False, is_aggregate_size=True):
         """
         List incomplete uploads list all previously uploaded incomplete multipart objects.
 
@@ -1201,8 +1088,8 @@ class Minio(object):
 
         # Initialize query parameters.
         query = {
-            'uploads': None,
-            'max-uploads': 1000
+            'uploads': '',
+            'max-uploads': '1000'
         }
 
         if prefix:
@@ -1221,16 +1108,19 @@ class Minio(object):
             response = self._url_open('GET',
                                       bucket_name=bucket_name,
                                       query=query)
-            (uploads, is_truncated, key_marker, upload_id_marker) = parse_list_multipart_uploads(response.data,
-                                                                                                 bucket_name=bucket_name)
+            (uploads, is_truncated, key_marker,
+             upload_id_marker) = parse_list_multipart_uploads(response.data,
+                                                              bucket_name=bucket_name)
             for upload in uploads:
                 if is_aggregate_size:
-                    upload.size = self._get_total_multipart_upload_size(upload.bucket_name,
-                                                                        upload.object_name,
-                                                                        upload.upload_id)
+                    upload.size = self._get_total_multipart_upload_size(
+                        upload.bucket_name,
+                        upload.object_name,
+                        upload.upload_id)
                 yield upload
 
-    def _get_total_multipart_upload_size(self, bucket_name, object_name, upload_id):
+    def _get_total_multipart_upload_size(self, bucket_name, object_name,
+                                         upload_id):
         """
         Get total multipart upload size.
 
@@ -1257,14 +1147,14 @@ class Minio(object):
 
         query = {
             'uploadId': upload_id,
-            'max-parts': 1000
+            'max-parts': '1000'
         }
 
         is_truncated = True
         part_number_marker = None
         while is_truncated:
             if part_number_marker:
-                query['part-number-marker'] = part_number_marker
+                query['part-number-marker'] = str(part_number_marker)
 
             response = self._url_open('GET',
                                       bucket_name=bucket_name,
@@ -1356,7 +1246,7 @@ class Minio(object):
         is_non_empty_string(object_name)
 
         if expires.total_seconds() < 1 or \
-           expires.total_seconds() > _SEVEN_DAYS_SECONDS:
+                        expires.total_seconds() > _SEVEN_DAYS_SECONDS:
             raise InvalidArgumentError('Expires param valid values'
                                        ' are between 1 secs to'
                                        ' {0} secs'.format(_SEVEN_DAYS_SECONDS))
@@ -1400,8 +1290,10 @@ class Minio(object):
                                                        date, region)
 
         post_policy.policies.append(('eq', '$x-amz-date', iso8601_date))
-        post_policy.policies.append(('eq', '$x-amz-algorithm', _SIGN_V4_ALGORITHM))
-        post_policy.policies.append(('eq', '$x-amz-credential', credential_string))
+        post_policy.policies.append(
+            ('eq', '$x-amz-algorithm', _SIGN_V4_ALGORITHM))
+        post_policy.policies.append(
+            ('eq', '$x-amz-credential', credential_string))
 
         post_policy_base64 = post_policy.base64()
         signature = post_presign_signature(date, region,
@@ -1421,7 +1313,7 @@ class Minio(object):
 
     # All private functions below.
     def _get_partial_object(self, bucket_name, object_name,
-                            offset=0, length=0):
+                            offset=0, length=0, request_headers=None):
         """Retrieves an object from a bucket.
 
         Optionally takes an offset and length of data to retrieve.
@@ -1442,6 +1334,8 @@ class Minio(object):
            Must be >= 0.
         :param length: Optional number of bytes to retrieve.
            Must be > 0.
+        :param request_headers: Any additional metadata to be uploaded along
+            with request.
         :return: :class:`urllib3.response.HTTPResponse` object.
 
         """
@@ -1449,6 +1343,9 @@ class Minio(object):
         is_non_empty_string(object_name)
 
         headers = {}
+        if request_headers:
+            headers = request_headers
+
         if offset != 0 or length != 0:
             request_range = '{}-{}'.format(
                 offset, "" if length == 0 else offset + length - 1
@@ -1504,7 +1401,7 @@ class Minio(object):
         return presign_url
 
     def _do_put_multipart_object(self, bucket_name, object_name, part_metadata,
-                                 upload_id='', part_number=0):
+                                 upload_id='', part_number=0, metadata=None):
         """
         Initiate a multipart PUT operation for a part number.
 
@@ -1513,26 +1410,36 @@ class Minio(object):
         :param part_metadata: Part-data and metadata for the multipart request.
         :param upload_id: Upload id of the multipart request.
         :param part_number: Part number of the data to be uploaded.
+        :param metadata: Any additional metadata to be uploaded along
+           with your object.
         """
         is_valid_bucket_name(bucket_name)
         is_non_empty_string(object_name)
         data = part_metadata.data
         if not callable(getattr(data, 'read')):
-            raise ValueError('Invalid input data does not implement a callable read() method')
+            raise ValueError(
+                'Invalid input data does not implement a callable read() method')
+
+        # Convert hex representation of md5 content to base64
+        md5content_b64 = codecs.encode(codecs.decode(
+            part_metadata.md5_hex, 'hex_codec'), 'base64_codec').strip()
 
         headers = {
             'Content-Length': part_metadata.size,
-            'Content-Md5': part_metadata.md5hasher.base64digest()
+            'Content-Md5': md5content_b64.decode()
         }
 
+        if metadata:
+            headers.update(metadata)
+
         response = self._url_open(
             'PUT', bucket_name=bucket_name,
             object_name=object_name,
             query={'uploadId': upload_id,
-                   'partNumber': part_number},
+                   'partNumber': str(part_number)},
             headers=headers,
             body=data,
-            content_sha256=part_metadata.sha256hasher.hexdigest()
+            content_sha256=part_metadata.sha256_hex
         )
 
         return response.headers['etag'].replace('"', '')
@@ -1585,28 +1492,25 @@ class Minio(object):
         :param content_size: Total size of the content to be uploaded.
         :param content_type: Content type of of the multipart upload.
            Defaults to 'application/octet-stream'.
+        :param metadata: Any additional metadata to be uploaded along
+           with your object.
         """
         is_valid_bucket_name(bucket_name)
         is_non_empty_string(object_name)
         if not callable(getattr(data, 'read')):
-            raise ValueError('Invalid input data does not implement a callable read() method')
+            raise ValueError(
+                'Invalid input data does not implement a callable read() method')
 
         # get upload id.
-        upload_id = self._get_upload_id(bucket_name, object_name, metadata)
+        upload_id = self._new_multipart_upload(bucket_name, object_name, metadata)
 
         # Initialize variables
         total_uploaded = 0
+        uploaded_parts = {}
 
         # Calculate optimal part info.
-        total_parts_count, part_size, last_part_size = optimal_part_info(content_size)
-
-        # Iter over the uploaded parts.
-        parts_iter = self._list_object_parts(bucket_name,
-                                             object_name,
-                                             upload_id)
-
-        # save uploaded parts for verification.
-        uploaded_parts = { part.part_number: part for part in parts_iter }
+        total_parts_count, part_size, last_part_size = optimal_part_info(
+            content_size)
 
         # Generate new parts and upload <= current_part_size until
         # part_number reaches total_parts_count calculated for the
@@ -1617,15 +1521,6 @@ class Minio(object):
                                  else last_part_size)
 
             part_metadata = parts_manager(data, current_part_size)
-            md5_hex = part_metadata.md5hasher.hexdigest()
-
-            # Verify if part number has been already uploaded.
-            # Further verify if we have matching md5sum as well.
-            previous_part = uploaded_parts.get(part_number, None)
-            if (previous_part and previous_part.size == current_part_size and
-                previous_part.etag == md5_hex):
-                total_uploaded += previous_part.size
-                continue
 
             # Seek back to starting position.
             part_metadata.data.seek(0)
@@ -1633,7 +1528,8 @@ class Minio(object):
                                                  object_name,
                                                  part_metadata,
                                                  upload_id,
-                                                 part_number)
+                                                 part_number,
+                                                 metadata=metadata)
             # Save etags.
             uploaded_parts[part_number] = UploadPart(bucket_name,
                                                      object_name,
@@ -1651,10 +1547,11 @@ class Minio(object):
             raise InvalidSizeError(msg)
 
         # Complete all multipart transactions if possible.
-        mpart_result =  self._complete_multipart_upload(bucket_name,
-                                                        object_name,
-                                                        upload_id,
-                                                        uploaded_parts)
+        mpart_result = self._complete_multipart_upload(bucket_name,
+                                                       object_name,
+                                                       upload_id,
+                                                       uploaded_parts,
+                                                       metadata=metadata)
         # Return etag here.
         return mpart_result.etag
 
@@ -1688,13 +1585,13 @@ class Minio(object):
 
         response = self._url_open('POST', bucket_name=bucket_name,
                                   object_name=object_name,
-                                  query={'uploads': None},
+                                  query={'uploads': ''},
                                   headers=metadata)
 
         return parse_new_multipart_upload(response.data)
 
     def _complete_multipart_upload(self, bucket_name, object_name,
-                                   upload_id, uploaded_parts):
+                                   upload_id, uploaded_parts, metadata=None):
         """
         Complete an active multipart upload request.
 
@@ -1702,6 +1599,8 @@ class Minio(object):
         :param object_name: Object name of the multipart request.
         :param upload_id: Upload id of the active multipart request.
         :param uploaded_parts: Key, Value dictionary of uploaded parts.
+        :param metadata: Any additional metadata to be uploaded along
+           with your object.
         """
         is_valid_bucket_name(bucket_name)
         is_non_empty_string(object_name)
@@ -1716,6 +1615,8 @@ class Minio(object):
             'Content-Type': 'application/xml',
             'Content-Md5': md5_base64,
         }
+        if metadata:
+            headers.update(metadata)
 
         response = self._url_open('POST', bucket_name=bucket_name,
                                   object_name=object_name,
@@ -1757,7 +1658,7 @@ class Minio(object):
             return self._region
 
         # get bucket location for Amazon S3.
-        region = 'us-east-1' # default to US standard.
+        region = 'us-east-1'  # default to US standard.
         if bucket_name in self._region_map:
             region = self._region_map[bucket_name]
         else:
@@ -1775,7 +1676,7 @@ class Minio(object):
         :return: location of bucket name is returned.
         """
         method = 'GET'
-        url = self._endpoint_url + '/' + bucket_name + '?location'
+        url = self._endpoint_url + '/' + bucket_name + '?location='
         headers = {}
         # default for all requests.
         region = 'us-east-1'
@@ -1798,8 +1699,7 @@ class Minio(object):
                       self._trace_output_stream)
 
         if response.status != 200:
-            response_error = ResponseError(response)
-            raise response_error.get(bucket_name)
+            raise ResponseError(response, method, bucket_name).get_exception()
 
         location = parse_location_constraint(response.data)
         # location is empty for 'US standard region'
@@ -1848,30 +1748,30 @@ class Minio(object):
             dump_http(method, url, fold_case_headers, response,
                       self._trace_output_stream)
 
-        if response.status != 200 and \
-           response.status != 204 and response.status != 206:
+        if response.status != 200 and response.status != 204 \
+           and response.status != 206:
             # Upon any response error invalidate the region cache
             # proactively for the bucket name.
             self._delete_bucket_region(bucket_name)
 
-            # Populate response_error with error response.
-            response_error = ResponseError(response)
-
             # In case we did not preload_content, we need to release
             # the connection:
             if not preload_content:
                 response.release_conn()
 
-            if method == 'HEAD':
-                raise response_error.head(bucket_name, object_name)
-            elif method == 'GET':
-                raise response_error.get(bucket_name, object_name)
-            elif method == 'POST':
-                raise response_error.post(bucket_name, object_name)
-            elif method == 'PUT':
-                raise response_error.put(bucket_name, object_name)
-            elif method == 'DELETE':
-                raise response_error.delete(bucket_name, object_name)
+            supported_methods = [
+                'HEAD',
+                'GET',
+                'POST',
+                'PUT',
+                'DELETE'
+            ]
+
+            if method in supported_methods:
+                raise ResponseError(response,
+                                    method,
+                                    bucket_name,
+                                    object_name).get_exception()
             else:
                 raise ValueError('Unsupported method returned'
                                  ' error: {0}'.format(response.status))
