@@ -35,7 +35,8 @@ import os
 import errno
 import math
 
-from .compat import urlsplit, urlencode, str, bytes, basestring
+from .compat import (urlsplit, urlencode, queryencode,
+                     str, bytes, basestring)
 from .error import (InvalidBucketError, InvalidEndpointError,
                     InvalidArgumentError)
 
@@ -117,14 +118,14 @@ class PartMetadata(object):
     Parts manager split parts metadata :class:`PartMetadata <PartMetadata>`.
 
     :param data: Part writer object backed by temporary file.
-    :param md5hasher: Computed md5sum Hasher interface.
-    :param sha256hasher: Computed sha256sum Hasher interface.
+    :param md5_hex: Md5 hash in hex format.
+    :param sha256_hex: Sha256 hash in hex format.
     :param size: Size of the part.
     """
-    def __init__(self, data, md5hasher, sha256hasher, size):
+    def __init__(self, data, md5_hex, sha256_hex, size):
         self.data = data
-        self.md5hasher = md5hasher
-        self.sha256hasher = sha256hasher
+        self.md5_hex = md5_hex
+        self.sha256_hex = sha256_hex
         self.size = size
 
 
@@ -149,52 +150,7 @@ def parts_manager(data, part_size=5*1024*1024):
         sha256hasher.update(current_data)
         total_read += len(current_data)
 
-    return PartMetadata(tmpdata, md5hasher, sha256hasher, total_read)
-
-
-def ignore_headers(headers_to_sign):
-    """
-    Ignore headers.
-    """
-    # Excerpts from @lsegal -
-    # https://github.com/aws/aws-sdk-js/issues/659#issuecomment-120477258
-    #
-    #  User-Agent:
-    #
-    #      This is ignored from signing because signing this causes problems
-    #      with generating pre-signed URLs (that are executed by other agents)
-    #      or when customers pass requests through proxies, which may modify
-    #      the user-agent.
-    #
-    #  Content-Length:
-    #
-    #      This is ignored from signing because generating a pre-signed URL
-    #      should not provide a content-length constraint, specifically when
-    #      vending a S3 pre-signed PUT URL. The corollary to this is that when
-    #      sending regular requests (non-pre-signed), the signature contains
-    #      a checksum of the body, which implicitly validates the payload
-    #      length (since changing the number of bytes would change the
-    #      checksum) and therefore this header is not valuable in the
-    #      signature.
-    #
-    #  Content-Type:
-    #
-    #      Signing this header causes quite a number of problems in browser
-    #      environments, where browsers like to modify and normalize the
-    #      content-type header in different ways. There is more information
-    #      on this in https://github.com/aws/aws-sdk-js/issues/244. Avoiding
-    #      this field simplifies logic and reduces the possibility of bugs.
-    #
-    #  Authorization:
-    #
-    #      Is skipped for obvious reasons
-    ignored_headers = ['Authorization', 'Content-Length',
-                       'Content-Type', 'User-Agent']
-    for ignored_header in ignored_headers:
-        if ignored_header in headers_to_sign:
-            del headers_to_sign[ignored_header]
-
-    return headers_to_sign
+    return PartMetadata(tmpdata, md5hasher.hexdigest(), sha256hasher.hexdigest(), total_read)
 
 AWS_S3_ENDPOINT_MAP = {
     'us-east-1': 's3.amazonaws.com',
@@ -235,10 +191,16 @@ def get_target_url(endpoint_url, bucket_name=None, object_name=None,
     parsed_url = urlsplit(endpoint_url)
 
     # Get new host, scheme.
+    scheme = parsed_url.scheme
     host = parsed_url.netloc
+
+    # Strip 80/443 ports since curl & browsers do not
+    # send them in Host header.
+    if parsed_url.port == 80 or parsed_url.port == 443:
+        host = parsed_url.hostname
+
     if 's3.amazonaws.com' in host:
         host = get_s3_endpoint(bucket_region)
-    scheme = parsed_url.scheme
 
     url = scheme + '://' + host
     if bucket_name:
@@ -265,13 +227,15 @@ def get_target_url(endpoint_url, bucket_name=None, object_name=None,
             if ordered_query[component_key] is not None:
                 if isinstance(ordered_query[component_key], list):
                     for value in ordered_query[component_key]:
-                        encoded_query = urlencode(
-                            str(value)).replace('/', '%2F')
-                        query_components.append(component_key+'='+encoded_query)
+                        query_components.append(component_key+'='+
+                                                queryencode(value))
                 else:
-                    encoded_query = urlencode(
-                        str(ordered_query[component_key])).replace('/', '%2F')
-                    query_components.append(component_key+'='+encoded_query)
+                    query_components.append(
+                        component_key+'='+
+                        queryencode(
+                            ordered_query[component_key]
+                        )
+                    )
             else:
                 query_components.append(component_key)
 
@@ -429,36 +393,31 @@ def is_valid_bucket_notification_config(notifications):
     for key, value in notifications.items():
         # check if key names are valid
         if key not in VALID_NOTIFICATION_KEYS:
-            raise InvalidArgumentError(
-                ('{} is an invalid key '
-                 'for notifications configuration').format(
-                     key
-                )
-            )
+            raise InvalidArgumentError((
+                '{} is an invalid key '
+                'for notifications configuration').format(key))
 
         # check if config values conform
         # first check if value is a list
         if not isinstance(value, list):
-            raise InvalidArgumentError(
-                ('The value for key "{}" in the notifications '
-                 'configuration must be a list.').format(key)
-            )
+            raise InvalidArgumentError((
+                'The value for key "{}" in the notifications '
+                'configuration must be a list.').format(key))
+
         for service_config in value:
             # check type matches
             if not isinstance(service_config, dict):
-                raise InvalidArgumentError(
-                    ('Each service configuration item for "{}" must be a '
-                     'dictionary').format(key)
-                )
+                raise InvalidArgumentError((
+                    'Each service configuration item for "{}" must be a '
+                    'dictionary').format(key))
+
             # check keys are valid
             for skey in service_config.keys():
                 if skey not in VALID_SERVICE_CONFIG_KEYS:
-                    raise InvalidArgumentError(
-                        ('{} is an invalid key for a service '
-                         'configuration item').format(
-                             skey
-                         )
-                    )
+                    raise InvalidArgumentError((
+                        '{} is an invalid key for a service '
+                        'configuration item').format(skey))
+
             # check for required keys
             arn = service_config.get('Arn', '')
             if arn == '':
@@ -472,22 +431,19 @@ def is_valid_bucket_notification_config(notifications):
                     'At least one event must be specified in a service config'
                 )
             if not isinstance(events, list):
-                raise InvalidArgumentError(
-                    '"Events" must be a list of strings in a service '
-                    'configuration'
-                )
+                raise InvalidArgumentError('"Events" must be a list of strings '
+                                           'in a service configuration')
+
             # check if 'Id' key is present, it should be string or bytes.
             if not isinstance(service_config.get('Id', ''), basestring):
-                raise InvalidArgumentError(
-                    '"Id" key must be a string'
-                )
+                raise InvalidArgumentError('"Id" key must be a string')
+
             for event in events:
                 if event not in NOTIFICATION_EVENTS:
                     raise InvalidArgumentError(
-                        '{} is not a valid event. Valid events are: {}'.format(
-                            event, NOTIFICATION_EVENTS
-                        )
-                    )
+                        '{} is not a valid event. Valid '
+                        'events are: {}'.format(event, NOTIFICATION_EVENTS))
+
             if 'Filter' in service_config:
                 exception_msg = (
                     '{} - If a Filter key is given, it must be a '
@@ -511,15 +467,14 @@ def is_valid_bucket_notification_config(notifications):
                         name = filter_rule['Name']
                         value = filter_rule['Value']
                     except KeyError:
-                        raise InvalidArgumentError(
-                            ('{} - a FilterRule dictionary must have "Name" '
-                             'and "Value" keys').format(filter_rule)
-                        )
+                        raise InvalidArgumentError((
+                            '{} - a FilterRule dictionary must have "Name" '
+                             'and "Value" keys').format(filter_rule))
+
                     if name not in ['prefix', 'suffix']:
-                        raise InvalidArgumentError(
-                            ('{} - The "Name" key in a filter rule must be '
-                             'either "prefix" or "suffix"').format(name)
-                        )
+                        raise InvalidArgumentError((
+                            '{} - The "Name" key in a filter rule must be '
+                             'either "prefix" or "suffix"').format(name))
 
     return True
 
